print("Loading wheel_assembly.lua")

-- The wheel assembly handles physical simulation and rendering of a wheel + suspension

WheelAssembly = WheelAssembly or { }

function WheelAssembly.new (is_left, obj, parent, wheel_mesh,
                            pos, wheel_radius, len,
                            steer, mass, mu, 
                            down, forwards, damping_factor)
        local self = {}

        -- these guys set and maintained later
        self.onGround = false
        self.skidding = 0
        self.locked = false
        self.tractionControl = 1
        self.travel = 0 -- gets converted to rotation
        self.freeWheelSpeed = 0
        self.forwards = Vector3(0,0,0)
        self.right = Vector3(0,0,0)
        self.restitutionForce = Vector3(0,0,0)
        self.antiRestitutionForce = Vector3(0,0,0)
        self.tractionForce = Vector3(0,0,0)
        self.frictionForce = Vector3(0,0,0)
        self.velocityDifference = Vector3(0,0,0)
        self.worldContactPos = Vector3(0,0,0)
        self.worldContactNormal = Vector3(0,0,0)

        local hooke = 2 * (9.81*mass) / len
        local damping = 2 * math.sqrt(mass * hooke) * (damping_factor or 0.5)
        self.suspensionLength = len
        self.suspensionHooke = hooke
        self.suspensionDamping = damping
        self.mu = mu
        self.obj = obj
        self.steerFactor = steer
        self.wheelRadius = wheel_radius or 0
        self.wheelCircumference = 2 * self.wheelRadius * math.pi
        self.pos = pos
        self.down = (down or V_DOWN).normalised
        self.forwardsBeforeSteer = (forwards or V_FORWARDS).normalised
        self.lastExtension = -self.suspensionLength/2
        make_instance(self,WheelAssembly)
        self:setSteer(0)
        self:setTorque(0)

        -- graphics
        self.rootNode = parent:createChild()
        self.rootNode.position = self.pos
        self.rootNode.orientation = V_FORWARDS:getRotationTo(self.down)
        -- wheel node does both the 'suspension' offset and 'steer' rotation
        self.wheelNode = self.rootNode:createChild()
        -- for spinning due to the engine
        self.wheelSpinNode = self.wheelNode:createChild()
        -- so that the wheels on the left hand side of the car are inverted
        self.wheelExtremeNode = self.wheelSpinNode:createChild()
        if is_left then
                self.wheelExtremeNode.orientation = Quat(180,V_UP)
        end
        self.wheelEntity = gfx.scnmgr:createEntity("Wheel:"..fresh(),wheel_mesh)
        self.wheelExtremeNode:attachObject(self.wheelEntity)
        self:updateGFX()
        return self
end

function WheelAssembly:destroy ()
        self.rootNode = safe_destroy(self.rootNode)
        self.wheelNode = safe_destroy(self.wheelNode)
        self.wheelSpinNode = safe_destroy(self.wheelSpinNode)
        self.wheelExtremeNode = safe_destroy(self.wheelExtremeNode)
        self.wheelEntity = safe_destroy(self.wheelEntity)
end

function WheelAssembly:setSteer(amount)
        self.steer = amount * self.steerFactor
        self.forwards:xyz(self.forwardsBeforeSteer)
        self.forwards:rotate(self.steer, self.down)
        self.right:setAsCross(self.forwards,self.down)
end

function WheelAssembly:setTorque(torque)
        self.driveForce = self.wheelRadius==0 and 0 or (torque / self.wheelRadius)
end

-- called from update callback, NOT step callback
function WheelAssembly:updateGFX()
        self.wheelNode:setPosition(0, self.lastExtension + self.suspensionLength,0)
        self.wheelNode:setOrientationAngle(self.steer,0,1,0)
        local angle = 360 * self.travel / self.wheelCircumference
        self.wheelSpinNode:setOrientationAngle(angle, -1, 0, 0)
end

function WheelAssembly:extensionDelta(extension)
        local extension_delta = extension - self.lastExtension
        self.lastExtension = extension
        return extension_delta
end

function WheelAssembly:freeWheel ()
        if math.abs(self.driveForce) > 10 and not self.locked and self.obj.engine.gear ~= 0 then
                self.freeWheelSpeed = 15 * self.driveForce/math.abs(self.driveForce)
        else
                -- wind down
                self.freeWheelSpeed = self.freeWheelSpeed*math.pow(0.7,self.obj.world.stepSize)
        end
        self.travel = (self.travel + self.freeWheelSpeed * self.obj.world.stepSize) % self.wheelCircumference
end

function WheelAssembly:groundWheel (distance)
        self.travel = (self.travel + distance) % self.wheelCircumference
        self.freeWheelSpeed = distance / self.obj.world.stepSize
end

function WheelAssembly:process ()

        local body = self.obj.body
        local world = self.obj.world

        -- _ws means world space

        -- ALLOCATION ALLOCATION
        local ray_reach, ground, normal =
                body:cast(-1, self.pos,
                          self.down,
                          self.suspensionLength + self.wheelRadius,
                          true, false, body)

        self.onGround = ray_reach ~= nil

        ray_reach = ray_reach or self.suspensionLength + self.wheelRadius

        self.worldContactPos:xyz(self.down)
        self.worldContactPos:scale(ray_reach)
        self.worldContactPos:append(self.pos)
        body:getLocalPosition(self.worldContactPos, self.worldContactPos)

        local extension = ray_reach - self.wheelRadius - self.suspensionLength

        extension = clamp(extension, -self.suspensionLength-self.wheelRadius, 0)


        local extension_rate = self:extensionDelta(extension) / world.stepSize

        self.skidding = 0

        if not self.onGround then
                self:freeWheel()
                return
        end


        self.worldContactNormal:xyz(normal)

        local spring_force_mag   = - extension * self.suspensionHooke

        local damping_force_mag  = - extension_rate * self.suspensionDamping


        -- this is the important one
        local surface_force_mag = spring_force_mag + damping_force_mag


        self.restitutionForce:xyz(self.worldContactNormal)
        self.restitutionForce:scale(surface_force_mag)
        --if player_ctrl.vehicle ~= nil and player_ctrl.vehicle.instance == self.obj then
        --        echo("RF: "..tostring(self.restitutionForce))
        --end
        body:force(self.restitutionForce, self.worldContactPos)

        -- subtract ground velocity from body velocity
        self.velocityDifference:xyz(ground:getLocalVelocity(self.worldContactPos,true))
        self.velocityDifference:scale(-1)
        self.velocityDifference:append(body:getLocalVelocity(self.worldContactPos,true))


        if self.locked then

                self.velocityDifference:removeComponent(self.worldContactNormal)

                local speed_difference = #self.velocityDifference

                local friction, drive = global_friction_model(self, surface_force_mag,
                                                              speed_difference, 0, 1)

                self.tractionForce:xyz(0,0,0)

                self.skidding = clamp((speed_difference-5)/35,0,1)

                self.frictionForce:xyz(self.velocityDifference)
                self.frictionForce:scale(1 / speed_difference) -- normalise
                self.frictionForce:scale(-friction)

        else

                -- divide into 'with roll' component and 'against roll' component
                -- ALLOCATION ALLOCATION
                local forward_ws = body.worldOrientation * self.forwards
                local forward_speed = self.velocityDifference:dot(forward_ws)
                -- ALLOCATION ALLOCATION
                local right_ws = body.worldOrientation * self.right
                local right_speed = self.velocityDifference:dot(right_ws)
                --local right_sign = sign(right_speed)
                local right_sign = sign(right_speed)


                local friction, drive, spin = global_friction_model(self, surface_force_mag,
                                                                    math.abs(right_speed),
                                                                    self.driveForce, self.tractionControl)

                self.skidding = spin and 1 or 0

                if self.skidding < 1 then
                        self:groundWheel(forward_speed * world.stepSize)
                else
                        self:freeWheel()
                end

                self.tractionForce:xyz(forward_ws)
                self.tractionForce:scale(drive)

                self.frictionForce:xyz(right_ws)
                self.frictionForce:scale(-1*right_sign*friction)

                --echo("T:"..tostring(self.tractionForce).."  F:"..tostring(self.frictionForce))
        end     

        body:force(self.frictionForce, self.worldContactPos)
        body:force(self.tractionForce, self.worldContactPos)

        self.antiRestitutionForce:xyz(self.frictionForce)
        self.antiRestitutionForce:append(self.tractionForce)
        self.antiRestitutionForce:append(self.restitutionForce)

        -- hack to avoid applying extremely large forces
        local magnitude = math.min(#self.antiRestitutionForce, ground.mass * 10)
        self.antiRestitutionForce:normalise()
        self.antiRestitutionForce:scale(-magnitude)
        ground:force(self.antiRestitutionForce, self.worldContactPos)

end

-- vim: shiftwidth=8:tabstop=8:expandtab
